#include <cmath>
#include <QDebug>
#include <iostream>

#include "pulsegenerator.h"

#define LTC26xx
//#define MCP4821

PulseGenerator::PulseGenerator()
{
    this->quit = false;

    bcm2835_init();
    if( bcm2835_spi_begin() == 0) {
        printf("\nL'initialisation du bus SPI a échoué. Avez-vous pensé à exécuter le programme avec `sudo` ?\n");
        exit(-1);
    } else {
        bcm2835_spi_setClockDivider(BCM2835_SPI_CLOCK_DIVIDER_65536);
        /*Le chip select à utiliser est déterminé dans la méthode setDacChannel() */
        bcm2835_spi_setChipSelectPolarity(CS_LD_DAC_1, LOW);
        bcm2835_spi_setChipSelectPolarity(CS_LD_DAC_1, LOW);
        bcm2835_spi_setDataMode(BCM2835_SPI_MODE0); // CPOL=0 => SCLK à 0 à l'état repos, CPHA=0 => donnée transmise au 1ier front d'horloge
    }

#ifdef LTC26xx
    this->vRef = 5.0;
#endif
#ifdef MCP4821
    this->vRef = 2.048;
#endif
}

/**
 * @brief PulseGenerator::~PulseGenerator
 */
PulseGenerator::~PulseGenerator() {
    bcm2835_spi_end();
    bcm2835_close();
}

/**
 * @brief PulseGenerator::generatorThread
 */
void PulseGenerator::generatorThread() {
    char buf[3];
    int i = 0;
    int nbDotsForRisingEdge = this->risindEdge.size();
    int nbDotsForFallingEdge = this->fallingEdge.size();
    unsigned char cmd = 0x30 | this->channel;

    qDebug() << "N° de canal : " << this->channel << endl;

    while( ! this->quit ) {
#ifdef LTC26xx
        // Génération du niveau bas
        if(this->tLow > 0) {
            buf[0] = cmd;
            buf[1] = ((this->vLow >> 8) & 0x00ff);
            buf[2] = this->vLow & 0x00ff;
            bcm2835_spi_writenb(buf, 3);
            //qDebug() << "Vlow : " << this->vLow << endl;
            bcm2835_delay(this->tLow);
	        //std::this_thread::sleep_for(std::chrono::microseconds{ this->tLow });
        }

        // Génération du front montant
        i = 0;
        while(i < nbDotsForRisingEdge) {
            buf[0] = cmd;
            buf[1] = ((this->risindEdge[ i ] >> 8) & 0x00ff);
            buf[2] = this->risindEdge[ i ] & 0x00ff;
            bcm2835_spi_writenb(buf, 3);
            i++;
            bcm2835_delayMicroseconds(100);
	        //std::this_thread::sleep_for(std::chrono::microseconds{ 100 });
        }

        // Génération du niveau haut
        if(this->tHigh > 0) {
            buf[0] = cmd;
            buf[1] = ((this->vHigh >> 8) & 0x00ff);
            buf[2] = this->vHigh & 0x00ff;
            bcm2835_spi_writenb(buf, 3);
            //qDebug() << "Vhigh : " << (int)buf[1] << ", " << int(buf[2]) << endl;
            bcm2835_delay(this->tHigh);
  	         //std::this_thread::sleep_for(std::chrono::microseconds{ this->tHigh });
        }

        // Génération du front descendant
        i = 0;
        while(i < nbDotsForFallingEdge) {
            buf[0] = cmd;
            buf[1] = ((this->fallingEdge[ i ] >> 8) & 0x00ff);
            buf[2] = this->fallingEdge[ i ] & 0x00ff;
            bcm2835_spi_writenb(buf, 3);
            i++;
            bcm2835_delayMicroseconds(100);
 	         //std::this_thread::sleep_for(std::chrono::microseconds{ 100 });
        }
#endif

#ifdef MCP4821
        // Génération du niveau bas
        buf[0] = 0x30 | ((this->vLow >> 8) & 0x000f);
        buf[1] = this->vLow & 0x00ff;
        bcm2835_spi_writenb(buf, 2);
        //qDebug() << "Vlow : " << this->vLow << endl;
        bcm2835_delay(this->tLow);

        // Génération du front montant
        i = 0;
        while(i < risindEdge.size()) {
            buf[0] = 0x30 | ((this->risindEdge[ i ] >> 8) & 0x000f);
            buf[1] = this->risindEdge[ i ] & 0x00ff;
            bcm2835_spi_writenb(buf, 2);
            i = (i+1) % nbDotsForRisingEdge;
            bcm2835_delayMicroseconds(100);
        }

        // Génération du niveau haut
        buf[0] = 0x30 | ((this->vHigh >> 8) & 0x000f);
        buf[1] = this->vHigh & 0x00ff;
        bcm2835_spi_writenb(buf, 2);
        //qDebug() << "Vhigh : " << this->vHigh << endl;
        bcm2835_delay(this->tHigh);

        // Génération du front descendant
        i = 0;
        while(i < nbDotsForFallingEdge) {
            buf[0] = 0x30 | ((this->fallingEdge[ i ] >> 8) & 0x000f);
            buf[1] = this->fallingEdge[ i ] & 0x00ff;
            bcm2835_spi_writenb(buf, 2);
            i = (i+1) % nbDotsForFallingEdge;
            bcm2835_delayMicroseconds(100);
        }
#endif
    }

    // Remettre à 0 le signal sur toutes les voies
#ifdef LTC26xx
    buf[0] = 0x3F;
    buf[1] = 0x00;
    buf[2] = 0x00;
    bcm2835_spi_writenb(buf, 3);
#endif
#ifdef MCP4821
        // Génération du niveau bas
        buf[0] = 0x30;
        buf[1] = 0x00;
        bcm2835_spi_writenb(buf, 2);
#endif
    return;
}

/**
 * Calcule les points d'une rampe à partir de son équation
 * @brief PulseGenerator::rasterizeWithLinearEquation
 * @param xi
 * @param yi
 * @param xf
 * @param yf
 * @param dots
 */
void PulseGenerator::rasterizeWithLinearEquation(u_int16_t xi, u_int16_t yi, u_int16_t xf, u_int16_t yf, std::vector<u_int16_t> * dots) {
      int x;
      u_int16_t y;
      double a,b ;
      a =(double) (yf-yi)/(xf-xi) ;
      b = yi - a * xi ;
      for ( x = xi ; x <= xf ; x++ ) {
        y = std::round(a * x + b) ;
#ifdef LTC26xx
        // Caler les valeurs de Vl et Vh sur le bit 16 quelle que soit la resolution (12, 14 ou 16)
        y = y << (16 - this->resolution);
#endif
        dots->push_back(y);
        //qDebug() << " y[" << dots->size() << "] = " << y << endl;
      }
}

/**
 * Calcule les points d'une rampe à partir de l'algorithme de Bresenham
 * @brief PulseGenerator::rasterizeWithBresenhamAlgorithm
 * @param xi
 * @param yi
 * @param xf
 * @param yf
 * @param dots
 */
void PulseGenerator::rasterizeWithBresenhamAlgorithm(u_int16_t xi, u_int16_t yi, u_int16_t xf, u_int16_t yf, std::vector<u_int16_t>& dots) {
    int dx
        ,dy
        ,i
        ,xinc
        ,yinc
        ,cumul
        ;
     u_int16_t x
        ,y
        ;

    x = xi ;
    y = yi ;
    dx = xf - xi ;
    dy = yf - yi ;
    xinc = ( dx > 0 ) ? 1 : -1 ; /* xinc tjrs à 1 dans notre situation */
    yinc = ( dy > 0 ) ? 1 : -1 ;
    dx = std::abs(dx) ;
    dy = std::abs(dy) ;

    //qDebug() << "dots[" << x << "] = " << y << endl;
    dots.push_back(y) ;

    if ( dx > dy ) {
        cumul = dx / 2 ;
        for ( i = 1 ; i <= dx ; i++ ) {
            x += xinc ;
            cumul += dy ;
            if ( cumul >= dx ) {
                cumul -= dx ;
                y += yinc ;
            }
            //qDebug() << "dots[" << x << "] = " << y << endl;
            dots.push_back(y);
        }
    } else {
        cumul = dy / 2 ;
        for ( i = 1 ; i <= dy ; i++ ) {
            y += yinc ;
            cumul += dx ;
            if ( cumul >= dy ) {
                cumul -= dy ;
                x += xinc ;
                //qDebug() << "dots[" << x << "] = " << y << endl;
                dots.push_back(y);
            }

        }
    }
}

/**
 * Détermine les valeurs à envoyer sur le DAC pour les rampes montantes et descendantes
 * @brief PulseGenerator::computeRisingEdge
 * @param Vi tension de départ
 * @param Vf tension d'arrivée
 * @param Tr temps de montée exprimé en ms
 */
void PulseGenerator::computeEdge(u_int16_t vi, u_int16_t vf, int t, edge_t edgeType) {
    std::vector<u_int16_t> * dots;
    u_int16_t xi = 0;
    u_int16_t xf = t*10; /* 't' étant la durée de la rampe en ms, on considère 1 point toutes les 100us (10pts/ms) */

    qDebug() << "vi = " << vi << endl;
    qDebug() << "vf = " << vf << endl;
    if(edgeType == RISING) {
        dots = &this->risindEdge;
    } else {
        dots = &this->fallingEdge;
    }
    dots->clear();

    auto start = std::chrono::system_clock::now();

    rasterizeWithLinearEquation(xi, vi, xf, vf, dots);
    
	 auto end = std::chrono::system_clock::now();
    auto elapsed = std::chrono::duration_cast<std::chrono::microseconds>(end - start);
    qDebug() << "Temps de calcul des points de la rampe : " << elapsed.count() << "us" << endl;
}

/**
 * Démarre la génération du signal DAC
 * @brief PulseGenerator::start
 */
void PulseGenerator::start() {
    this->quit = false;
    t = std::thread(&PulseGenerator::generatorThread, this);

    return;
}

/**
 * Stoppe la génération du signal DAC
 * @brief PulseGenerator::stop
 * @return
 */
int PulseGenerator::stop() {
    this->quit = true;
    if (t.joinable()) {
        t.join();
    }
    return 0;
}

/**
 * Mémorise la résolution du dAC et calcule la valeur pleine échelle
 * @brief PulseGenerator::setDacResolution
 * @param resolution
 */
void PulseGenerator::setDacResolution(int resolution) {
#ifdef MCP4821
    resolution = 12;
#endif

    // Mémoriser la résolution et la valeur pleine échelle
    this->resolution = resolution;
    this->fullScale = (1 << resolution) - 1;
}

/**
 * Sélectionne le DAC et mémorise le numéro de canal sur lequel le signal doit être généré
 * @brief PulseGenerator::setDacChannel
 * @param resolution
 */
void PulseGenerator::setDacChannel(int channel) {
    if (channel < 10) {
        // Sélectionner le DAC correspondant à la voie ([0...7]-> Dac1, [10...13] -> Dac2)
        bcm2835_spi_chipSelect(CS_LD_DAC_1);
        // Mémoriser le n° de voie du DAC sur lequel le signal sera généré
        this->channel = channel;
    } else {
        // Sélectionner le DAC correspondant à la voie ([0...7]-> Dac1, [10...13] -> Dac2)
        bcm2835_spi_chipSelect(CS_LD_DAC_2);
        // Mémoriser le n° de voie du DAC sur lequel le signal sera généré
        this->channel = channel - 10;
    }
}

/**
 * Configure la fréquence de Sclck dans la librairie bcm2835 en fonction du choix de l'utilisateur
 * @brief PulseGenerator::setSclck
 * @param div
 */
void PulseGenerator::setSclck(int div) {
    int spiClockDivider;

    switch(div) {
    case 8 :
        spiClockDivider = BCM2835_SPI_CLOCK_DIVIDER_8;
        break;
    case 16 :
        spiClockDivider = BCM2835_SPI_CLOCK_DIVIDER_16;
        break;
    case 32 :
        spiClockDivider = BCM2835_SPI_CLOCK_DIVIDER_32;
        break;
    case 64 :
        spiClockDivider = BCM2835_SPI_CLOCK_DIVIDER_64;
        break;
    default :
        spiClockDivider = BCM2835_SPI_CLOCK_DIVIDER_8;
        break;
    }

    bcm2835_spi_setClockDivider(spiClockDivider);
}

/**
 * Détermine tous les points du signal DAC (rampes et niveaux)
 * @brief PulseGenerator::computeSignal
 * @param tl
 * @param tr
 * @param th
 * @param tf
 * @param vl
 * @param vh
 */
void PulseGenerator::computeSignal(int tl, int tr, int th, int tf, double vl, double vh) {
    this->vLow = std::round(vl * this->fullScale / this->vRef);
    this->vHigh = std::round(vh * this->fullScale / this->vRef);

    this->tLow = tl;
    this->tHigh = th;

    if(tr) {
        this->computeEdge(this->vLow, this->vHigh, tr, RISING);
    } else {
        this->risindEdge.clear();
    }

    if(tf) {
        this->computeEdge(this->vHigh, this->vLow, tf, FALLING);
    } else {
        this->fallingEdge.clear();
    }

#ifdef LTC26xx
    // Caler les valeurs de Vl et Vh sur le bit 16 quelle que soit la resolution (12, 14 ou 16)
    this->vLow = this->vLow << (16 - this->resolution);
    this->vHigh = this->vHigh << (16 - this->resolution);
#endif

    return;
}

